/*
 * jlibinit.c
 *
 * Copyright 2011 Tom Gall <tom.gall@linaro.org> for Linaro Limited
 *
 * This file is for library initialization
 * Alg notes:
 *   At the time the library is loaded, libjpeg_general_init is called
 *     this will run before main, in the case that the library is NOT
 *     loaded via dlopen
 *   libjpeg_general_init will call libjpeg_arch_specific_init which
 *     if there is work to do should be implemented by each architecture
 *     otherwise the empty stub will be called.
 * 
 *   Example: In the case of arm, we want to query the auxv to 
 *   determine if neon hardware is present. Further if present
 *   functional pointers can be setup such that the simd versions
 *   will be used.
 */

#include <link.h>
#include <elf.h>
#include <malloc.h>
#include <unistd.h>
#include <stdio.h>

#include "jinclude.h"
#include "jpeglib.h"
#include "jsimd.h"

#if defined(__linux__) || defined(__APPLE__)
LOCAL(void) libjpeg_arch_specific_init(void);

GLOBAL(void __attribute__ ((constructor))) libjpeg_general_init(void);

#if !defined(__APPLE__)
volatile ElfW(auxv_t) *auxv = NULL;

LOCAL(ElfW(auxv_t) *)get_auxv(void)
{
  FILE *auxv_f;
  ElfW(auxv_t) auxv_struct;
  int i = 0;

  if(auxv == NULL) {
    auxv_f = fopen("/proc/self/auxv", "r");

    if(auxv_f == 0) {
       perror("Error opening file for reading");
          return 0;
    }
    auxv =(ElfW(auxv_t) *)malloc(getpagesize());

    do
      {
      fread(&auxv_struct, sizeof(ElfW(auxv_t)), 1, auxv_f);
      auxv[i] = auxv_struct;
      i++;
      } while(auxv_struct.a_type != AT_NULL);
  }
  return auxv;
}
#else

/* For OSes without an auxv implementation
 */
LOCAL(ElfW(auxv_t) *)get_auxv(void)
{
}

#endif /* if not on apple */

/* 
 * Runs immediately after library load (and before main() unless
 * dlopen is used.
 */
GLOBAL(void __attribute__ ((constructor))) libjpeg_general_init(void) 
{
  // Architecture independent library init

  get_auxv();

  // call to arch specific init
  libjpeg_arch_specific_init();
}

#if defined(__arm__) || defined(__i386__)

LOCAL(void) libjpeg_arch_specific_init(void) 
{

  init_simd();

}
#elif defined(__x86_64__)

LOCAL(void) libjpeg_arch_specific_init(void) 
{

}

#else

/*
 * empty stub for architectures that don't define an arch specific
 * init routine.
 */
LOCAL(void) libjpeg_arch_specific_init(void) 
{

}

#endif /* __arch__ */



#elif defined(_WIN32) /* Windows */

GLOBAL (BOOL) WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason,
      LPVOID lpReserved)
{
  switch(fdwReason) {
    case DLL_PROCESS_ATTACH:
      // call to arch specific init
      libjpeg_arch_specific_init();
      break
    case DLL_PROCESS_DETACH:
      break
  }

  return (TRUE);
}

#if defined(__i386__)

LOCAL(void) libjpeg_arch_specific_init(void) 
{

  init_simd();

}
#elif defined(__x86_64__)

LOCAL(void) libjpeg_arch_specific_init(void) 
{

}
#endif /* arch for Windows */
#endif /* __linux__ __APPLE__  */
